/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* dynamic memory allocation for GNU.
   Copyright (C) 1985, 1987 Free Software Foundation, Inc. */

/*
 * @(#)nmalloc.c 1 (Caltech) 2/21/82
 *
 *	U of M Modified: 20 Jun 1983 ACT: strange hacks for Emacs
 *
 *	Nov 1983, Mike@BRL, Added support for 4.1C/4.2 BSD.
 *
 * This is a very fast storage allocator.  It allocates blocks of a small 
 * number of different sizes, and keeps free lists of each size.  Blocks
 * that don't exactly fit are passed up to the next larger size.  In this 
 * implementation, the available sizes are (2^n)-4 (or -16) bytes long.
 * This is designed for use in a program that uses vast quantities of
 * memory, but bombs when it runs out.  To make it a little better, it
 * warns the user when he starts to get near the end.
 *
 * June 84, ACT: modified rcheck code to check the range given to malloc,
 * rather than the range determined by the 2-power used.
 *
 * Jan 85, RMS: calls malloc_warning to issue warning on nearly full.
 * No longer Emacs-specific; can serve as all-purpose malloc for GNU.
 * You should call malloc_init to reinitialize after loading dumped Emacs.
 * Call malloc_stats to get info on memory stats if MSTATS turned on.
 * realloc knows how to return same block given, just changing its size,
 * if the power of 2 is correct.
 *
 * Jul 90, MW: ported to Amiga, using Intuition's AllocRemember-Feature
 * to keep track of alloc'd data. I threw out all Unix-code (sbrk..), so
 * this a true Amiga version.
 */

#define KERNEL
#include "ixemul.h"
#include <exec/memory.h>

#undef DEBUG

#ifdef DEBUG
#define ADD_KALLOC(a) add_kalloc(a)
#define REM_KALLOC(a) rem_kalloc(a)
#else
#define ADD_KALLOC(a)
#define REM_KALLOC(a)
#endif

#define rcheck

/* Amiga-specific stuff... 
 * The big difference to the Unix version, is, that you can't tell in
 * advance, how much memory you'll get next time, since there's only one
 * memory pool.. */ 

#define BASE_EXT_DECL
#define BASE_PAR_DECL	
#define BASE_PAR_DECL0	
#define BASE_NAME	ix.ix_intui_base
#include <inline/intuition.h>

#define botch(str) ix_panic (str), syscall (SYS_exit, 20)

/*
 * nextf[i] is the pointer to the next free block of size 2^(i+3).  The
 * smallest allocatable block is 8 bytes.  The overhead information will
 * go in the first int of the block, and the returned pointer will point
 * to the second.
 */

/* this is true for the current configuration of the MMU-Tables.. however,
 * it's almost a don't care on the Amiga... */
#define PAGESIZE 256

#define ISALLOC ((char) 0xf7)	/* magic byte that implies allocation */
#define ISFREE ((char) 0x54)	/* magic byte that implies free block */
				/* this is for error checking only */
#define ISMEMALIGN ((char) 0xd6)  /* Stored before the value returned by
				     memalign, with the rest of the word
				     being the distance to the true
				     beginning of the block.  */

/* This is for user programs to look at, when they are interested.  */

#define malloc_sbrk_used	(md->md_malloc_sbrk_used)    /* amount of data space used now */

/* use this to get a chunk of the specified size */
static inline void *
get_chunk (struct malloc_data *md, int size)
{
  malloc_sbrk_used += size;
  return AllocRemember(& md->md_key, size, MEMF_PUBLIC);
}


/* If range checking is not turned on, all we have is a flag indicating
   whether memory is allocated, an index in nextf[], and a size field; to
   realloc() memory we copy either size bytes or 1<<(index+3) bytes depending
   on whether the former can hold the exact size (given the value of
   'index').  If range checking is on, we always need to know how much space
   is allocated, so the 'size' field is never used. */

struct mhead {
	char     mh_alloc;	/* ISALLOC or ISFREE */
	char     mh_index;	/* index in nextf[] */
/* Remainder are valid only when block is allocated */
	unsigned short mh_size;	/* size, if < 0x10000 */
#ifdef rcheck
	unsigned mh_nbytes;	/* number of bytes allocated */
	int      mh_magic4;	/* should be == MAGIC4 */
#endif /* rcheck */
};

/* Access free-list pointer of a block.
  It is stored at block + 4.
  This is not a field in the mhead structure
  because we want sizeof (struct mhead)
  to describe the overhead for when the block is in use,
  and we do not want the free-list pointer to count in that.  */

#define CHAIN(a) \
  (*(struct mhead **) (sizeof (char *) + (char *) (a)))

#ifdef rcheck
/* To implement range checking, we write magic values in at the beginning and
   end of each allocated block, and make sure they are undisturbed whenever a
   free or a realloc occurs. */
/* Written in each of the 4 bytes following the block's real space */
#define MAGIC1 0x55
/* Written in the 4 bytes before the block's real space */
#define MAGIC4 0x55555555
#define ASSERT(p) if (!(p)) botch(#p); else
#define EXTRA  4		/* 4 bytes extra for MAGIC1s */
#else
#define ASSERT(p)
#define EXTRA  0
#endif /* rcheck */

/* nextf[i] is free list of blocks of size 2**(i + 3)  */

#define nextf (md->md_nextf)
/* static struct mhead *nextf[30]; */

/* busy[i] is nonzero while allocation of block size i is in progress.  */

#define busy (md->md_busy)
/* static char busy[30]; */

/* nonzero once initial bunch of free blocks made */

#define gotpool (md->md_gotpool)
/* static int gotpool; */

static void getpool (struct malloc_data *md);
static void *sys_malloc (struct malloc_data *md, size_t size);
static void *sys_realloc (struct malloc_data *md, void *old, size_t size);
static void *sys_memalign (struct malloc_data *md, size_t align, size_t size);
static void sys_free (struct malloc_data *md, void *buf);

/* Return the maximum size to which MEM can be realloc'd
   without actually requiring copying.  */

static int
malloc_usable_size (void *mem)
{
  int blocksize = 8 << (((struct mhead *) mem) - 1) -> mh_index;

  return blocksize - sizeof (struct mhead) - EXTRA;
}

static void
morecore (struct malloc_data *md, int nu)	/* ask system for more memory */
{
  register char *cp;
  register int nblks;
  register unsigned int siz;

 /* On initial startup, get two blocks of each size up to 1k bytes */
  if (!gotpool)
    { getpool (md); getpool (md); gotpool = 1; }

 /* Take at least 2k, and figure out how many blocks of the desired size
    we're about to get */
  nblks = 1;
  if ((siz = nu) < 8)
    nblks = 1 << ((siz = 8) - nu);

  if (!(cp = get_chunk(md, 1 << (siz + 3))))
    return;			/* no more room! */

 /* save new header and link the nblks blocks together */
  nextf[nu] = (struct mhead *) cp;
  siz = 1 << (nu + 3);
  while (1)
    {
      ((struct mhead *) cp) -> mh_alloc = ISFREE;
      ((struct mhead *) cp) -> mh_index = nu;
      if (--nblks <= 0) break;
      CHAIN ((struct mhead *) cp) = (struct mhead *) (cp + siz);
      cp += siz;
    }
  CHAIN ((struct mhead *) cp) = 0;

}

static void
getpool (struct malloc_data *md)
{
  register int nu;
  char *cp;

  /* Get 2k of storage */

  cp = get_chunk (md, 04000);
  if (cp == (char *) -1)
    return;

  /* Divide it into an initial 8-word block
     plus one block of size 2**nu for nu = 3 ... 10.  */

  CHAIN (cp) = nextf[0];
  nextf[0] = (struct mhead *) cp;
  ((struct mhead *) cp) -> mh_alloc = ISFREE;
  ((struct mhead *) cp) -> mh_index = 0;
  cp += 8;

  for (nu = 0; nu < 7; nu++)
    {
      CHAIN (cp) = nextf[nu];
      nextf[nu] = (struct mhead *) cp;
      ((struct mhead *) cp) -> mh_alloc = ISFREE;
      ((struct mhead *) cp) -> mh_index = nu;
      cp += 8 << nu;
    }
}

static void *
sys_malloc (struct malloc_data *md, size_t n)
{
  register struct mhead *p;
  register unsigned int nbytes;
  register int nunits = 0;

  /* Figure out how many bytes are required, rounding up to the nearest
     multiple of 4, then figure out which nextf[] area to use */
  nbytes = (n + sizeof *p + EXTRA + 3) & ~3;
  {
    register unsigned int   shiftr = (nbytes - 1) >> 2;

    while (shiftr >>= 1)
      nunits++;
  }

  /* In case this is reentrant use of malloc from signal handler,
     pick a block size that no other malloc level is currently
     trying to allocate.  That's the easiest harmless way not to
     interfere with the other level of execution.  */
  while (busy[nunits]) nunits++;
  busy[nunits] = 1;

  /* If there are no blocks of the appropriate size, go get some */
  /* COULD SPLIT UP A LARGER BLOCK HERE ... ACT */
  if (nextf[nunits] == 0)
    morecore (md, nunits);

  /* Get one block off the list, and set the new list head */
  if ((p = nextf[nunits]) == 0)
    {
      busy[nunits] = 0;
      return 0;
    }
  nextf[nunits] = CHAIN (p);
  busy[nunits] = 0;

  /* Check for free block clobbered */
  /* If not for this check, we would gobble a clobbered free chain ptr */
  /* and bomb out on the NEXT allocate of this size block */
  if (p -> mh_alloc != ISFREE || p -> mh_index != nunits)
#ifdef rcheck
    botch ("block on free list clobbered");
#else /* not rcheck */
    syscall (SYS_abort);
#endif /* not rcheck */

  /* Fill in the info, and if range checking, set up the magic numbers */
  p -> mh_alloc = ISALLOC;
#ifdef rcheck
  p -> mh_nbytes = n;
  p -> mh_magic4 = MAGIC4;
  {
    register char  *m = (char *) (p + 1) + n;

    *m++ = MAGIC1, *m++ = MAGIC1, *m++ = MAGIC1, *m = MAGIC1;
  }
#else /* not rcheck */
  p -> mh_size = n;
#endif /* not rcheck */
  return (void *) (p + 1);
}

static void
sys_free (struct malloc_data *md, void *mem)
{
  register struct mhead *p;
  {
    register char *ap = mem;
 
    if (ap == 0)
      return;

    p = (struct mhead *) ap - 1;

    if (p -> mh_alloc == ISMEMALIGN)
      {
	ap -= p->mh_nbytes;
	p = (struct mhead *) ap - 1;
      }

#ifndef rcheck
    if (p -> mh_alloc != ISALLOC)
      syscall (SYS_abort);

#else /* rcheck */
    if (p -> mh_alloc != ISALLOC)
      {
	if (p -> mh_alloc == ISFREE)
	  botch ("free: Called with already freed block argument\n");
	else
	  botch ("free: Called with bad argument\n");
      }

    ASSERT (p -> mh_magic4 == MAGIC4);
    ap += p -> mh_nbytes;
    ASSERT (*ap++ == MAGIC1); ASSERT (*ap++ == MAGIC1);
    ASSERT (*ap++ == MAGIC1); ASSERT (*ap   == MAGIC1);
#endif /* rcheck */
  }
  {
    register int nunits = p -> mh_index;

    ASSERT (nunits <= 29);
    p -> mh_alloc = ISFREE;

    /* Protect against signal handlers calling malloc.  */
    busy[nunits] = 1;
    /* Put this block on the free list.  */
    CHAIN (p) = nextf[nunits];
    nextf[nunits] = p;
    busy[nunits] = 0;

  }
}

static void *
sys_realloc (struct malloc_data *md, void *mem, size_t n)
{
  register struct mhead *p;
  register unsigned int tocopy;
  register unsigned int nbytes;
  register int nunits;

  if ((p = (struct mhead *) mem) == 0)
    return (void *) sys_malloc(md, n);
  p--;
  nunits = p -> mh_index;
  ASSERT (p -> mh_alloc == ISALLOC);
#ifdef rcheck
  ASSERT (p -> mh_magic4 == MAGIC4);
  {
    register char *m = mem + (tocopy = p -> mh_nbytes);
    ASSERT (*m++ == MAGIC1); ASSERT (*m++ == MAGIC1);
    ASSERT (*m++ == MAGIC1); ASSERT (*m   == MAGIC1);
  }
#else /* not rcheck */
  if (p -> mh_index >= 13)
    tocopy = (1 << (p -> mh_index + 3)) - sizeof *p;
  else
    tocopy = p -> mh_size;
#endif /* not rcheck */

  /* See if desired size rounds to same power of 2 as actual size. */
  nbytes = (n + sizeof *p + EXTRA + 7) & ~7;

  /* If ok, use the same block, just marking its size as changed.  */
  if (nbytes > (4 << nunits) && nbytes <= (8 << nunits))
    {
#ifdef rcheck
      register char *m = mem + tocopy;
      *m++ = 0;  *m++ = 0;  *m++ = 0;  *m++ = 0;
      p-> mh_nbytes = n;
      m = mem + n;
      *m++ = MAGIC1;  *m++ = MAGIC1;  *m++ = MAGIC1;  *m++ = MAGIC1;
#else /* not rcheck */
      p -> mh_size = n;
#endif /* not rcheck */
      return mem;
    }

  if (n < tocopy)
    tocopy = n;
  {
    register void *new;

    if ((new = (void *) sys_malloc(md, n)) == 0)
      return 0;
    bcopy (mem, new, tocopy);
    sys_free (md, mem);
    return new;
  }
}

static void *
sys_memalign (struct malloc_data *md, size_t alignment, size_t size)
{
  register char *ptr = sys_malloc (md, size + alignment);
  register char *aligned;
  register struct mhead *p;

  if (ptr == 0)
    return 0;
  /* If entire block has the desired alignment, just accept it.  */
  if (((int) ptr & (alignment - 1)) == 0)
    return ptr;
  /* Otherwise, get address of byte in the block that has that alignment.  */
  aligned = (char *) (((int) ptr + alignment - 1) & -alignment);

  /* Store a suitable indication of how to free the block,
     so that free can find the true beginning of it.  */
  p = (struct mhead *) aligned - 1;
  p -> mh_size = aligned - ptr;
  p -> mh_alloc = ISMEMALIGN;
  return aligned;
}

void *
malloc (size_t size)
{
  return sys_malloc (&u.u_md, size);
}

void *
realloc (void *mem, size_t size)
{
  return sys_realloc (&u.u_md, mem, size);
}

void *
memalign (size_t alignment, size_t size)
{
  return sys_memalign (&u.u_md, alignment, size);
}

void *
valloc (size_t size)
{
  return sys_memalign(&u.u_md, PAGESIZE, size);
}

void
free (void *mem)
{
  sys_free (&u.u_md, mem);
}

#if 1
void *
kmalloc (size_t size)
{
  void *res;

  res = sys_malloc (&ix.ix_md, size);
  ADD_KALLOC (res);
  return res;
}

void *
krealloc (void *mem, size_t size)
{
  void *res;

  REM_KALLOC (mem);
  res = sys_realloc (&ix.ix_md, mem, size);
  ADD_KALLOC (res);
  return res;
}

void *
kmemalign (size_t alignment, size_t size)
{
  void *res;

  res = sys_memalign (&ix.ix_md, alignment, size);
  ADD_KALLOC (res);
  return res;
}

void *
kvalloc (size_t size)
{
  void *res;

  res = sys_memalign(&ix.ix_md, PAGESIZE, size);
  ADD_KALLOC (res);
  return res;
}

void
kfree (void *mem)
{
  REM_KALLOC (mem);
  sys_free (&ix.ix_md, mem);
}
#else
void *
kmalloc (size_t size)
{
  return syscall (SYS_malloc, size);
}

void *
krealloc (void *mem, size_t size)
{
  return syscall (SYS_realloc, mem, size);
}

void *
kmemalign (size_t alignment, size_t size)
{
  return syscall (SYS_memalign, alignment, size);
}

void *
kvalloc (size_t size)
{
  return syscall (SYS_memalign, PAGESIZE, size);
}

void
kfree (void *mem)
{
  syscall (SYS_free, mem);
}
#endif


#ifdef DEBUG
static void *addr_tab[1000];

void
add_kalloc (void *adr)
{
  void **vp;
  
  for (vp = addr_tab; vp < &addr_tab[1000]; vp++)
    if (*vp == adr)
      {
        ix_panic ("add_kalloc: adr already in table!");
        return;
      }
  
  for (vp = addr_tab; vp < &addr_tab[1000]; vp++)
    if (! *vp) break;
    
  if (*vp)
    {
      ix_panic ("add_kalloc: table full!");
      return;
    }

  *vp = adr;
}

void
rem_kalloc (void *adr)
{
  void **vp;
  
  for (vp = addr_tab; vp < &addr_tab[1000]; vp++)
    if (*vp == adr)
      {
        *vp = 0;
        return;
      }

  ix_panic ("rem_kalloc: freeing not allocated memory!");
}
 
void
check_kalloc ()
{
  void **vp;

  for (vp = addr_tab; vp < &addr_tab[1000]; vp++)
    if (*vp)
      kprintf ("check_kalloc: found still allocated block at $%lx.\n", *vp);
}

void
dump_remember (struct malloc_data *md)
{
  struct Remember *r;

  kprintf ("dump_remember:\n");

  kprintf ("md_malloc_sbrk_used = %ld\n", md->md_malloc_sbrk_used);
  kprintf ("md_got_pool = %ld\n", md->md_gotpool);
  
  if (md->md_key)
    {
      for (r = md->md_key; r; r = r->NextRemember)
	kprintf ("  @$%lx: size = %ld, address = $%lx, next = $%lx.\n", r, r->RememberSize, r->Memory, r->NextRemember);
    }
  else
    kprintf ("  no remember data.\n");
}
#endif
